private import SDL;
private import SDL_image;
private import SDL_mixer;
private import image;
private import sound;
private import mouse;
private import tango.io.FilePath;
private import tango.io.Stdout;
private import tango.io.stream.TextFileStream;
private import tango.util.PathUtil;
private import tango.sys.Environment;
private import tango.text.convert.Integer;
private import tango.text.Regex;

version (Win32) {
private import tango.stdc.stringz;
private import tango.sys.win32.UserGdi;
private import tango.text.convert.Utf;
}

void main(char[][] args) {
  try {
    auto exeFile = args[0];
    version (Win32) {
      exeFile = normalize(exeFile);
    }
    auto exeFilePath = Environment.exePath(exeFile);
    auto main = new Main(exeFilePath);
    main.run;
  } catch (Exception e) {
    version (Win32) {
      // WINAPIを呼んでみるテスト
      char[] message = e.toString();
      wchar[] convertedMessage = Utf.toString16(message);
      MessageBoxW(null, toString16z(convertedMessage), "Error", MB_OK | MB_ICONEXCLAMATION);
    } else {
      Stdout(e.toString);
    }

    // キャッチした例外をそのままもう一度投げます
    throw e;
  }
}

private final class Main {
  private const FilePath exeFilePath;
  private Mouse mouse;
  private Screen screen;
  /// 殴る画像の配列
  private Image[] beatImages;
  /// 撫でる画像の配列
  private Image[] strokeImages;
  /// 0-9の数字が並んだ画像
  private Image numbersImage;
  private char[][char[]] configures;
  private Sound beatSound;
  private Sound laughSound;

  this(FilePath exeFilePath) {
    this.exeFilePath = exeFilePath;
  }

  void run() {
    init;
    load;
    loop;
    quit;
  }

  /// いろんな準備
  private void init() {
    if (SDL_Init(SDL_INIT_EVERYTHING) < 0) {
      throw new Exception("failed to initialize SDL");
    }
    if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 2, 1024) < 0) {
      throw new Exception("failed to initialize SDL_mixer");
    }
  }

  /// 実行ファイルからの相対パスを受け取って、カレントディレクトリからのパスとして解決して返します。
  private char[] resolveRelativePath(char[] path) {
    return exeFilePath.file(path).toString;
  }

  /// 設定や画像、音声の読み込み
  private void load() {
    //Properties!(char[]).load("config.txt", loadConfig); // tango.text.Preferences は deprecated なので不採用。
    auto configStream = new TextFileInput(resolveRelativePath("config.txt"));
    auto regex = Regex(r"^\s*([^;]?\w+)\s*=\s*(\S+)\s*$");
    // Regexとビルドオプション-debugを併用すると"Access Violation"が発生するので注意！
    foreach (char[] line;  configStream) {
      if (regex.test(line)) {
        auto key = regex.match(1);
        auto value = regex.match(2);
        configures[key] = value;
      }
    }

    uint windowWidth = toInt!(char)(configures["WINDOWWIDTH"]);
    assert(windowWidth > 0, "ウィンドウの横幅の設定値が異常です！");
    uint windowHeight = toInt!(char)(configures["WINDOWHEIGHT"]);
    assert(windowHeight > 0, "ウィンドウの縦幅の設定値が異常です！");

    screen = new Screen(windowWidth, windowHeight);
    if ("CAPTION" in configures) {
      screen.caption = configures["CAPTION"];
    }

    // 画像、音声などのロード
    beatImages ~= new Image(resolveRelativePath(configures["BEATIMAGE1"]));
    beatImages ~= new Image(resolveRelativePath(configures["BEATIMAGE2"]));
    beatImages ~= new Image(resolveRelativePath(configures["BEATIMAGE3"]));
    strokeImages ~= new Image(resolveRelativePath(configures["STROKEIMAGE1"]));
    numbersImage = new Image(resolveRelativePath(configures["NUMBERIMAGE"]));
    beatSound = new Sound(resolveRelativePath(configures["BEATSOUND"]));
    laughSound = new Sound(resolveRelativePath(configures["STROKESOUND"]));
    auto swingThreshold = screen.width * 1 / 3;
    mouse = new Mouse(swingThreshold);
  }

  /// メインの処理
  private void loop() {
    SDL_Event event;
    bool done = false;

    // 今殴っている（マウスを押している）かどうか
    bool beating = false;
    // 殴った回数
    uint beatCount = 0;
    // 振った回数
    uint swingCount = 0;

    while (!done) {
      uint ticks = SDL_GetTicks();
      while (SDL_PollEvent(&event)) {
        switch (event.type) {
        case SDL_QUIT:
          done = true;
          break;
        case SDL_MOUSEBUTTONDOWN:
          beatSound.play;
          beating = true;
          beatCount++;
          // クリックしたら撫でていた状態をリセットする
          mouse.click;
          swingCount = 0;
          break;
        case SDL_MOUSEBUTTONUP:
          beating = false;
          break;
        case SDL_MOUSEMOTION:
          // マウスカーソルがウィンドウ上にある時だけ発生する
          mouse.move(event.motion.x, event.motion.y);
          break;
        default:
        }
      }

      // ここから描画開始
      screen.clear;
      screen.drawImage(beating ? beatImages[1] : mouse.swingCount > 0 ? strokeImages[0] : beatCount == 0 ? beatImages[0] : beatImages[2]);
      if (mouse.swingCount > swingCount && !laughSound.playing) {
        // 撫で続けても最初の一回しか効果音を鳴らさないようにする
        swingCount = int.max;
        laughSound.play;
      }

      // 殴った回数を表示
      uint digitWidth = numbersImage.width/10;
      uint digitHeight = numbersImage.height;
      uint num = beatCount;
      for (int i=0;; ++i) {
        int digit = num%10;
        int x = screen.width - digitWidth*(i+1);
        int y = screen.height - digitHeight;
        screen.drawImage(numbersImage, digitWidth*digit, 0, digitWidth, digitHeight, x, y);
        num /= 10;
        if (num == 0) break;
      }

      // 画面の再描画
      if (!screen.flip()) {
        throw new Exception("failed to flip screen.");
      }

      // 1/60秒くらい待ちます
      SDL_Delay(1000/60);
    }
  }

  /// お片付け
  private void quit() {
    foreach (Image image; beatImages) {
      delete image;
    }
    foreach (Image image; strokeImages) {
      delete image;
    }
    delete numbersImage;
    delete beatSound;
    delete laughSound;

    SDL_Quit();
    Mix_CloseAudio();
  }
}